package izayoi.cache.impl.selene;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author es
 */
public class SeleneCompatibleCache {

    private File file;
    private Charset charset;
    private static final Pattern ptn = Pattern.compile("(\\d+)?([\\/|\\-]{1})?([0-9a-f]+)?");

    /**
     * @param file
     */
    public SeleneCompatibleCache(File file, Charset charset) {
        this.charset = charset;
        if (file == null) {
            throw new IllegalArgumentException("file is null");
        }
        this.file = file;
    }

    /**
     * @param string
     */
    public void append(String string) throws IOException {
        ChannelHolder ch = new ChannelHolder(file, "rw", charset);
        try {
            if (!ch.lock(false, 119)) {
                throw new IOException("busy");
            }
            ch.seek(file.length());
            ch.writeln(string);
        } finally {
            ch.close();
        }
    }

    /**
     * @param time
     * @param id
     * @return
     */
    public String getRecord(long time, String id) throws IOException {
        ChannelHolder ch = new ChannelHolder(file, "r", charset);
        String key = String.valueOf(time) + "<>";
        if (id != null) {
            key += id;
        }

        try {
            if (!ch.lock(true, 119)) {
                throw new IOException("busy");
            }

            String s = ch.readln();
            while (s != null) {
                if (s.startsWith(key)) {
                    return s;
                }
                s = ch.readln();
            }
        } finally {
            ch.close();
        }
        return null;
    }

    /**
     * @param l
     * @param string
     */
    public void deleteRecord(long time, String id) throws IOException {
        String target = String.valueOf(time) + "<>" + id;
        ChannelHolder ch = new ChannelHolder(file, "rw", charset);
        try {
            if (!ch.lock(false, 119)) {
                throw new IOException("busy");
            }

            long pos = ch.position();
            String line = ch.readln();

            while (line != null) {
                if (line.startsWith(target)) {
                    break;
                }
                pos = ch.position();
                line = ch.readln();
            }

            List l = new ArrayList();
            line = ch.readln();
            while (line != null) {
                l.add(line);
                line = ch.readln();
            }

            ch.seek(pos);
            for (Iterator i = l.iterator(); i.hasNext();) {
                ch.writeln((String) i.next());
            }

        } finally {
            ch.close();
        }
    }

    /**
     * 
     */
    public File getFile() {
        return file;
    }

    /**
     * @return
     */
    public String getLastRecord() throws IOException {
        ChannelHolder ch = new ChannelHolder(file, "r", charset);
        try {
            if (!ch.lock(true, 119)) {
                throw new IOException("busy");
            }
            String ret = null;
            String s = ch.readln();
            while (s != null) {
                ret = s;
                s = ch.readln();
            }
            return ret;
        } finally {
            ch.close();
        }
    }

    /**
     * @return
     */
    public int getRecordCount() throws IOException {
        return getRecords().length;
    }

    /**
     * @return
     */
    public String[] getRecords() throws IOException {
        ChannelHolder ch = new ChannelHolder(file, "r", charset);
        List l = new ArrayList();
        try {
            if (!ch.lock(true, 119)) {
                throw new IOException("busy");
            }
            String s = ch.readln();
            while (s != null) {
                l.add(s);
                s = ch.readln();
            }
        } finally {
            ch.close();
        }

        return (String[]) l.toArray(new String[l.size()]);
    }

    /**
     * @param argument
     * @return
     */
    public String[] searchRecord(String argument) throws NumberFormatException, IOException {
        List l = new ArrayList();

        Matcher matcher = ptn.matcher(argument);
        if (matcher.find()) {
            String time = matcher.group(1);
            String sp = matcher.group(2);
            String id = matcher.group(3);
            if (sp == null || sp.equals("/")) {
                if (time != null) {
                    String s = getRecord(Long.valueOf(time).longValue(), id);
                    if (s != null) {
                        l.add(s);
                    }
                }
            } else {
                if (!(time == null && id == null)) {
                    List r = getRecords(time, id);
                    if (r != null) {
                        l.addAll(r);
                    }
                }
            }
        }

        return (String[]) l.toArray(new String[l.size()]);
    }

    protected List getRecords(String from, String to) throws IOException {
        if (from == null) {
            from = "0000000000";
        }
        if (to == null) {
            to = "9999999999";
        }

        String[] records = getRecords();
        Comparator c = new Finder();
        Arrays.sort(records, c);

        String tmp = records[0];
        tmp = tmp.substring(0, tmp.indexOf("<"));
        if (tmp.compareTo(to) > 0) {
            return null;
        }
        tmp = records[records.length - 1];
        tmp = tmp.substring(0, tmp.indexOf("<"));
        if (tmp.compareTo(from) < 0) {
            return null;
        }

        int start = getPosition(records, from, c, true);
        int end = getPosition(records, to, c, false);
        return Arrays.asList(records).subList(start, end + 1);
    }

    /**
     * @param records
     * @param from
     * @param c
     * @return
     */
    protected int getPosition(String[] records, String time, Comparator c, boolean isUnder) {
        int min = 0;
        int max = records.length;
        int pos = (min + max) / 2;
        boolean isBreak = false;
        while (!isBreak) {
            int i = c.compare(time, records[pos]);
            if (i == 0) {
                isBreak = true;
            } else {
                if (i < 0) {
                    max = pos;
                } else {
                    min = pos;
                }
                int tmp = (min + max) / 2;
                if (tmp == pos) {
                    isBreak = true;
                    if (isUnder && records[pos].substring(0, records[pos].indexOf("<")).compareTo(time) < 0) {
                        pos++;
                        if (pos > records.length) {
                            pos = records.length;
                        }
                    }
                } else {
                    pos = tmp;
                }
            }
        }
        return pos;
    }

    /**
     * @param time
     * @param id
     * @return
     */
    public boolean containsRecord(long time, String id) throws IOException {
        return (getRecord(time, id) != null);
    }

}

class Finder implements Comparator {

    /* ( Javadoc)
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public int compare(Object arg0, Object arg1) {
        String s1 = (String) arg0;
        String s2 = (String) arg1;

        if (s1.indexOf("<") != -1) {
            s1 = s1.substring(0, s1.indexOf("<"));
        }
        if (s2.indexOf("<") != -1) {
            s2 = s2.substring(0, s2.indexOf("<"));
        }
        return s1.compareTo(s2);
    }
}

class ChannelHolder {
    private String charset;
    private FileChannel ch;
    private FileLock fl;
    private int BUFFER_SIZE = 4096 * 8;
    private byte[] array = new byte[BUFFER_SIZE];
    private ByteBuffer buff = ByteBuffer.wrap(array);

    ChannelHolder(File target, String mode, Charset charset) {
        this.charset = charset.name();
        try {
            if (!target.exists()) {
                target.createNewFile();
            }
            ch = new RandomAccessFile(target, mode).getChannel();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    void close() throws IOException {
        ch.close();
        unLock();
    }

    boolean lock(boolean isShear, int challenge) throws IOException {
        if (fl != null && fl.isValid()) {
            return false;
        }

        for (int i = 0; i < challenge; i++) {
            fl = ch.tryLock(0, Long.MAX_VALUE, isShear);
            if (fl != null) {
                break;
            }
            try {
                Thread.sleep(500);
            } catch (InterruptedException e) {
                e.printStackTrace(System.err);
                break;
            }
        }
        return (fl != null);
    }

    String readln() throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        buff.position(0);

        boolean isCR = false;
        while (!isCR) {
            int count = ch.read(buff);
            if (count <= 0) {
                isCR = true;
                continue;
            }

            int j = 0;
            for (; j < count; j++) {
                if (array[j] == 10) {
                    isCR = true;
                    break;
                }
            }

            out.write(array, 0, j);
            if (j != count) {
                long l = ch.position() - (count - j) + 1;
                ch.position(l);
            }
            buff.position(0);
        }

        String ret = null;
        if (out.size() > 0) {
            ret = new String(out.toByteArray(), charset).trim();
        }
        return ret;
    }

    void unLock() throws IOException {
        if (fl != null && fl.isValid()) {
            fl.release();
        }
    }

    void writeln(String string) throws IOException {
        string += "\r\n";
        ch.write(ByteBuffer.wrap(string.getBytes(charset)));
        ch.truncate(ch.position());
    }

    void seek(long l) throws IOException {
        ch.position(l);
    }

    long position() throws IOException {
        return ch.position();
    }

}
